@@ -3,7 +3,7 @@
 /*                  This file is part of the class library                   */
 /*       SoPlex --- the Sequential object-oriented simPlex.                  */
 /*                                                                           */
-/*  Copyright (c) 1996-2023 Zuse Institute Berlin (ZIB)                      */
+/*  Copyright (c) 1996-2025 Zuse Institute Berlin (ZIB)                      */
 /*                                                                           */
 /*  Licensed under the Apache License, Version 2.0 (the "License");          */
 /*  you may not use this file except in compliance with the License.         */
@@ -163,6 +163,18 @@ public:
    /**@name Inquiry */
    ///@{
 
+   /// returns current tolerances
+   const std::shared_ptr<Tolerances> tolerances() const
+   {
+      return _tolerances;
+   }
+
+   /// set tolerances
+   virtual void setTolerances(std::shared_ptr<Tolerances> tolerances)
+   {
+      this->_tolerances = tolerances;
+   }
+
    /// Returns true if and only if the LP is scaled
    bool isScaled() const
    {
@@ -626,10 +638,10 @@ public:
    void addRow(const S* lhsValue, const S* rowValues, const int* rowIndices, int rowSize,
                const S* rhsValue)
    {
-      assert(lhsValue != 0);
-      assert(rowSize <= 0 || rowValues != 0);
-      assert(rowSize <= 0 || rowIndices != 0);
-      assert(rhsValue != 0);
+      assert(lhsValue != nullptr);
+      assert(rowSize <= 0 || rowValues != nullptr);
+      assert(rowSize <= 0 || rowIndices != nullptr);
+      assert(rhsValue != nullptr);
 
       int idx = nRows();
       int oldColNumber = nCols();
@@ -677,12 +689,12 @@ public:
    void addRows(const S* lhsValues, const S* rowValues, const int* rowIndices, const int* rowStarts,
                 const int* rowLengths, const int numRows, const int numValues, const S* rhsValues)
    {
-      assert(lhsValues != 0);
-      assert(numValues <= 0 || rowValues != 0);
-      assert(numValues <= 0 || rowIndices != 0);
-      assert(numValues <= 0 || rowStarts != 0);
-      assert(numValues <= 0 || rowLengths != 0);
-      assert(rhsValues != 0);
+      assert(lhsValues != nullptr);
+      assert(numValues <= 0 || rowValues != nullptr);
+      assert(numValues <= 0 || rowIndices != nullptr);
+      assert(numValues <= 0 || rowStarts != nullptr);
+      assert(numValues <= 0 || rowLengths != nullptr);
+      assert(rhsValues != nullptr);
 
       int i, j, k, idx;
       SVectorBase<R>* col;
@@ -697,7 +709,7 @@ public:
          assert(numValues <= 0 || rowStarts[i] + rowLengths[i] <= numValues);
 
          if(numValues <= 0)
-            LPRowSetBase<R>::add(&(lhsValues[i]), (S*)0, (int*)0, 0, &(rhsValues[i]));
+            LPRowSetBase<R>::add(&(lhsValues[i]), (S*)0, (int*)nullptr, 0, &(rhsValues[i]));
          else
             LPRowSetBase<R>::add(&(lhsValues[i]), &(rowValues[rowStarts[i]]), &(rowIndices[rowStarts[i]]),
                                  rowLengths[i], &(rhsValues[i]));
@@ -848,12 +860,12 @@ public:
                 const int* colStarts, const int* colLengths, const int numCols, const int numValues,
                 const S* upperValues)
    {
-      assert(lowerValues != 0);
-      assert(numValues <= 0 || colValues != 0);
-      assert(numValues <= 0 || colIndices != 0);
-      assert(numValues <= 0 || colStarts != 0);
-      assert(numValues <= 0 || colLengths != 0);
-      assert(upperValues != 0);
+      assert(lowerValues != nullptr);
+      assert(numValues <= 0 || colValues != nullptr);
+      assert(numValues <= 0 || colIndices != nullptr);
+      assert(numValues <= 0 || colStarts != nullptr);
+      assert(numValues <= 0 || colLengths != nullptr);
+      assert(upperValues != nullptr);
 
       int i, j, k, idx;
       SVectorBase<R>* row;
@@ -869,7 +881,7 @@ public:
          assert(numValues <= 0 || colStarts[i] + colLengths[i] <= numValues);
 
          if(numValues <= 0)
-            LPColSetBase<R>::add(&(objValue[i]), &(lowerValues[i]), (S*)0, (int*)0, 0, &(upperValues[i]));
+            LPColSetBase<R>::add(&(objValue[i]), &(lowerValues[i]), (S*)0, (int*)nullptr, 0, &(upperValues[i]));
          else
             LPColSetBase<R>::add(&(objValue[i]), &(lowerValues[i]), &(colValues[colStarts[i]]),
                                  &(colIndices[colStarts[i]]), colLengths[i], &(upperValues[i]));
@@ -982,10 +994,10 @@ public:
    }
 
    ///
-   virtual void removeRows(SPxRowId id[], int n, int perm[] = 0)
+   virtual void removeRows(SPxRowId id[], int n, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          DataArray < int > p(nRows());
          removeRows(id, n, p.get_ptr());
@@ -1007,10 +1019,10 @@ public:
     *  permutations resulting from this removal: \p perm[i] < 0 indicates, that the element to index \p i has been
     *  removed.  Otherwise, \p perm[i] is the new index of the element with index \p i before the removal.
     */
-   virtual void removeRows(int nums[], int n, int perm[] = 0)
+   virtual void removeRows(int nums[], int n, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          DataArray < int > p(nRows());
          removeRows(nums, n, p.get_ptr());
@@ -1027,10 +1039,10 @@ public:
    }
 
    /// Removes rows from \p start to \p end (including both).
-   virtual void removeRowRange(int start, int end, int perm[] = 0)
+   virtual void removeRowRange(int start, int end, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          int i = end - start + 1;
          DataArray < int > p(i);
@@ -1082,10 +1094,10 @@ public:
    }
 
    ///
-   virtual void removeCols(SPxColId id[], int n, int perm[] = 0)
+   virtual void removeCols(SPxColId id[], int n, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          DataArray < int > p(nCols());
          removeCols(id, n, p.get_ptr());
@@ -1107,10 +1119,10 @@ public:
     *  permutations resulting from this removal: \p perm[i] < 0 indicates, that the element to index \p i has been
     *  removed.  Otherwise, \p perm[i] is the new index of the element with index \p i before the removal.
     */
-   virtual void removeCols(int nums[], int n, int perm[] = 0)
+   virtual void removeCols(int nums[], int n, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          DataArray < int > p(nCols());
          removeCols(nums, n, p.get_ptr());
@@ -1127,10 +1139,10 @@ public:
    }
 
    /// Removes columns from \p start to \p end (including both).
-   virtual void removeColRange(int start, int end, int perm[] = 0)
+   virtual void removeColRange(int start, int end, int perm[] = nullptr)
    {
 
-      if(perm == 0)
+      if(perm == nullptr)
       {
          int i = end - start + 1;
          DataArray < int > p(i);
@@ -1177,12 +1189,12 @@ public:
    ///@{
 
    /// Reads LP in LP format from input stream \p in.
-   virtual bool readLPF(std::istream& in, NameSet* rowNames = 0, NameSet* colNames = 0,
-                        DIdxSet* intVars = 0);
+   virtual bool readLPF(std::istream& in, NameSet* rowNames = nullptr, NameSet* colNames = nullptr,
+                        DIdxSet* intVars = nullptr);
 
    /// Reads an LP in MPS format from input stream \p in.
-   virtual bool readMPS(std::istream& in, NameSet* rowNames = 0, NameSet* colNames = 0,
-                        DIdxSet* intVars = 0);
+   virtual bool readMPS(std::istream& in, NameSet* rowNames = nullptr, NameSet* colNames = nullptr,
+                        DIdxSet* intVars = nullptr);
 
    /// Reads LP in LP or MPS format from input stream \p in.
    /**@param in       input stream.
@@ -1195,8 +1207,8 @@ public:
     *                 file.  Maybe 0 if the information is not needed.
     * @todo Make sure the Id's in the NameSet%s are the same as in the LP.
     */
-   virtual bool read(std::istream& in, NameSet* rowNames = 0, NameSet* colNames = 0,
-                     DIdxSet* intVars  = 0)
+   virtual bool read(std::istream& in, NameSet* rowNames = nullptr, NameSet* colNames = nullptr,
+                     DIdxSet* intVars  = nullptr)
    {
       bool ok;
       char c;
@@ -1216,8 +1228,9 @@ public:
    }
 
    /// Reads LP from a file.
-   virtual bool readFile(const char* filename, NameSet* rowNames = 0, NameSet* colNames = 0,
-                         DIdxSet* intVars = 0)
+   virtual bool readFile(const char* filename, NameSet* rowNames = nullptr,
+                         NameSet* colNames = nullptr,
+                         DIdxSet* intVars = nullptr)
    {
 
       spxifstream file(filename);
@@ -1232,16 +1245,19 @@ public:
     *  constraints and variables. If \p intVars is not \c NULL, the variables contained in it are marked as integer in
     *  the output.
     */
-   virtual void writeLPF(std::ostream&  out, const NameSet* rowNames, const NameSet* colNames,
-                         const DIdxSet* p_intvars = 0) const;
+   virtual void writeLPF(std::ostream& out, const NameSet* rowNames, const NameSet* colNames,
+                         const DIdxSet* p_intvars = nullptr,
+                         const bool writeZeroObjective = false) const;
 
    /// Writes a file in MPS format to \p out.
-   virtual void writeMPS(std::ostream&  out, const NameSet* rowNames, const NameSet* colNames,
-                         const DIdxSet* p_intvars = 0) const;
+   virtual void writeMPS(std::ostream& out, const NameSet* rowNames, const NameSet* colNames,
+                         const DIdxSet* p_intvars = nullptr,
+                         const bool writeZeroObjective = false) const;
 
    /// Write loaded LP to \p filename.
-   virtual void writeFileLPBase(const char* filename, const NameSet* rowNames = 0,
-                                const NameSet* colNames = 0, const DIdxSet* p_intvars = 0) const
+   virtual void writeFileLPBase(const char* filename, const NameSet* rowNames = nullptr,
+                                const NameSet* colNames = nullptr, const DIdxSet* p_intvars = nullptr,
+                                const bool writeZeroObjective = false) const
    {
 
       std::ofstream tmp(filename);
@@ -1250,11 +1266,11 @@ public:
       if(len_f > 4 && filename[len_f - 1] == 's' && filename[len_f - 2] == 'p'
             && filename[len_f - 3] == 'm' && filename[len_f - 4] == '.')
       {
-         writeMPS(tmp, rowNames, colNames, p_intvars);
+         writeMPS(tmp, rowNames, colNames, p_intvars, writeZeroObjective);
       }
       else
       {
-         writeLPF(tmp, rowNames, colNames, p_intvars);
+         writeLPF(tmp, rowNames, colNames, p_intvars, writeZeroObjective);
       }
    }
 
@@ -1319,7 +1335,7 @@ public:
 
          if(hasRhs && hasLhs)
          {
-            if(EQ(lhs(i), rhs(i)))
+            if(EQ(lhs(i), rhs(i), this->tolerances()->epsilon()))
                countEqual++;
             else
                countRanged++;
@@ -1832,7 +1848,7 @@ public:
       SVectorBase<R>& row = rowVector_w(i);
       SVectorBase<R>& col = colVector_w(j);
 
-      if(isNotZero(val))
+      if(isNotZero(val, this->tolerances()->epsilon()))
       {
          R newVal;
 
@@ -1929,6 +1945,21 @@ public:
    virtual void computePrimalActivity(const VectorBase<R>& primal, VectorBase<R>& activity,
                                       const bool unscaled = true) const;
 
+   /// Computes activity of the \p i 'th row for a given primal vector
+   /// @throw SPxInternalCodeException if the dimension of primal vector does not match number of columns
+   /// \p unscaled determines whether the returned data should be unscaled (if scaling was applied prior)
+   virtual R computePrimalActivity(const int i, const VectorBase<R>& primal,
+                                   const bool unscaled = true) const;
+
+   /// Computes activities of the rows \p ids for a given primal vector; activities do not need to be zero;
+   /// only the elements with indices in \p ids are computed, the remaining elements of \p activity are not changed
+   /// @throw SPxInternalCodeException if the dimension of primal vector does not match number of columns or if the
+   ///        dimension of the activity vector does not match the number of rows
+   /// \p unscaled determines whether the returned data should be unscaled (if scaling was applied prior)
+   virtual void computePrimalActivity(const std::vector<int>& ids, const VectorBase<R>& primal,
+                                      VectorBase<R>& activity,
+                                      const bool unscaled = true) const;
+
    /// Updates activity of the rows for a given primal vector; activity does not need to be zero
    /// @throw SPxInternalCodeException if the dimension of primal vector does not match number of columns or if the
    ///        dimension of the activity vector does not match the number of rows
@@ -1948,11 +1979,27 @@ public:
    }
 
    /// Computes "dual" activity of the columns for a given dual vector, i.e., y^T A; activity does not need to be zero
-   /// @throw SPxInternalCodeException if dimension of dual vector does not match number of rows or if the dimension of
+   /// @throw SPxInternalCodeException if the dimension of dual vector does not match number of rows or if the dimension of
    ///        the activity vector does not match the number of columns
+   /// \p unscaled determines whether the returned data should be unscaled (if scaling was applied prior)
    virtual void computeDualActivity(const VectorBase<R>& dual, VectorBase<R>& activity,
                                     const bool unscaled = true) const;
 
+   /// Computes "dual" activity of the \p i 'th column for a given dual vector, i.e., y^T A
+   /// @throw SPxInternalCodeException if the dimension of dual vector does not match number of rows
+   /// \p unscaled determines whether the returned data should be unscaled (if scaling was applied prior)
+   virtual R computeDualActivity(const int i, const VectorBase<R>& dual,
+                                 const bool unscaled = true) const;
+
+   /// Computes "dual" activities of the columns \p ids for a given dual vector, i.e., y^T A; activities do not need to be zero;
+   /// only the elements with indices in \p ids are computed, the remaining elements of \p activity are not changed
+   /// @throw SPxInternalCodeException if the dimension of dual vector does not match number of rows or if the
+   ///        dimension of the activity vector does not match the number of columns
+   /// \p unscaled determines whether the returned data should be unscaled (if scaling was applied prior)
+   virtual void computeDualActivity(const std::vector<int>& ids, const VectorBase<R>& dual,
+                                    VectorBase<R>& activity,
+                                    const bool unscaled = true) const;
+
    /// Updates "dual" activity of the columns for a given dual vector, i.e., y^T A; activity does not need to be zero
    /// @throw SPxInternalCodeException if dimension of dual vector does not match number of rows or if the dimension of
    ///        the activity vector does not match the number of columns
@@ -2002,10 +2049,11 @@ public:
    /// Building the dual problem from a given LP
    /// @note primalRows must be as large as the number of unranged primal rows + 2 * the number of ranged primal rows.
    ///       dualCols must have the identical size to the primal rows.
-   virtual void buildDualProblem(SPxLPBase<R>& dualLP, SPxRowId primalRowIds[] = 0,
-                                 SPxColId primalColIds[] = 0,
-                                 SPxRowId dualRowIds[] = 0, SPxColId dualColIds[] = 0, int* nprimalrows = 0, int* nprimalcols = 0,
-                                 int* ndualrows = 0, int* ndualcols = 0);
+   virtual void buildDualProblem(SPxLPBase<R>& dualLP, SPxRowId primalRowIds[] = nullptr,
+                                 SPxColId primalColIds[] = nullptr,
+                                 SPxRowId dualRowIds[] = nullptr, SPxColId dualColIds[] = nullptr, int* nprimalrows = nullptr,
+                                 int* nprimalcols = nullptr,
+                                 int* ndualrows = nullptr, int* ndualcols = nullptr);
 
    ///@}
 
@@ -2016,6 +2064,9 @@ public:
    /// Consistency check.
    bool isConsistent() const
    {
+      if(this->_tolerances == nullptr && nCols() != 0)
+         return SPX_MSG_INCONSISTENT("SPxLPBase");
+
 #ifdef ENABLE_CONSISTENCY_CHECKS
 
       for(int i = nCols() - 1; i >= 0; --i)
@@ -2028,10 +2079,10 @@ public:
             int n = w.pos(i);
 
             if(n < 0)
-               return MSGinconsistent("SPxLPBase");
+               return SPX_MSG_INCONSISTENT("SPxLPBase");
 
             if(v.value(j) != w.value(n))
-               return MSGinconsistent("SPxLPBase");
+               return SPX_MSG_INCONSISTENT("SPxLPBase");
          }
       }
 
@@ -2045,10 +2096,10 @@ public:
             int n = w.pos(i);
 
             if(n < 0)
-               return MSGinconsistent("SPxLPBase");
+               return SPX_MSG_INCONSISTENT("SPxLPBase");
 
             if(v.value(j) != w.value(n))
-               return MSGinconsistent("SPxLPBase");
+               return SPX_MSG_INCONSISTENT("SPxLPBase");
          }
       }
 
@@ -2062,6 +2113,8 @@ public:
 
 protected:
 
+   std::shared_ptr<Tolerances> _tolerances;
+
    // ------------------------------------------------------------------------------------------------------------------
    /**@name Protected write access */
    ///@{
@@ -2806,7 +2859,7 @@ public:
    ///@{
 
    /// Default constructor.
-   SPxLPBase<R>()
+   SPxLPBase()
    {
       SPxLPBase<R>::clear(); // clear is virtual.
 
@@ -2814,11 +2867,11 @@ public:
    }
 
    /// Destructor.
-   virtual ~SPxLPBase<R>()
+   virtual ~SPxLPBase()
    {}
 
    /// Copy constructor.
-   SPxLPBase<R>(const SPxLPBase<R>& old)
+   SPxLPBase(const SPxLPBase<R>& old)
       : LPRowSetBase<R>(old)
       , LPColSetBase<R>(old)
       , thesense(old.thesense)
@@ -2827,12 +2880,13 @@ public:
       , lp_scaler(old.lp_scaler)
       , spxout(old.spxout)
    {
+      _tolerances = old._tolerances;
       assert(isConsistent());
    }
 
    /// Copy constructor.
    template < class S >
-   SPxLPBase<R>(const SPxLPBase<S>& old)
+   SPxLPBase(const SPxLPBase<S>& old)
       : LPRowSetBase<R>(old)
       , LPColSetBase<R>(old)
       , thesense(old.thesense == SPxLPBase<S>::MINIMIZE ? SPxLPBase<R>::MINIMIZE : SPxLPBase<R>::MAXIMIZE)
@@ -2841,6 +2895,7 @@ public:
       , spxout(old.spxout)
    {
       lp_scaler = nullptr;
+      _tolerances = old._tolerances;
       assert(isConsistent());
    }
 
@@ -2856,6 +2911,7 @@ public:
          _isScaled = old._isScaled;
          lp_scaler = old.lp_scaler;
          spxout = old.spxout;
+         _tolerances = old._tolerances;
 
          assert(isConsistent());
       }
@@ -2879,6 +2935,7 @@ public:
                     SPxLPBase<R>::MAXIMIZE;
          offset = R(old.offset);
          _isScaled = old._isScaled;
+         _tolerances = old._tolerances;
 
          // this may have un-intended consequences in the future
          lp_scaler = nullptr;
